﻿/*******************************************************************************
* Copyright (C) 2014 - 2016 Intel Corp. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Intel Corp. nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Intel Corp. OR THE CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/

using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Management;

namespace FwUpdateAPI
{
    /// <summary>
    /// This class is the interface to a host controller
    ///
    /// This, with its base class, is the main interface of this API module for applications
    /// to interface with host controllers.
    /// The main role of this class is to wrap the interface to WMI but it includes also the
    /// interface for validating image compatibility with the current controller.
    /// </summary>
    public class SdkTbtController : SdkTbtBase
    {
        /// <summary>
        /// C-tor
        /// </summary>
        /// <param name="mo">WMI object to use for interfacing with the controller</param>
        public SdkTbtController(ManagementObject mo) : base(mo)
        {
        }

        /// <summary>
        /// This property is used to differentiate between the power down messages
        /// given to user for host controller or device controller after update in
        /// cases where the PD version was updated. 
        /// </summary>
        /// <returns>Power down necessary for host controller message string</returns>
        public override string GetNeedPowerDownMessage
        {
            get { return Resources.HostNeedsPowerDownMessage; }
        }

        /// <summary>
        /// Controller ID property as returned by WMI
        /// </summary>
        public string ControllerId
        {
            get { return Mo.GetPropertyValue("ControllerId").ToString(); }
        }

        /// <summary>
        /// True if the controller is in safe-mode according to the information from WMI
        /// </summary>
        public bool IsInSafeMode
        {
            get { return (bool) Mo.GetPropertyValue("IsInSafeMode"); }
        }

        /// <summary>
        /// Gets list of SdkTbtController objects for each of the host controllers exposed by WMI
        /// </summary>
        /// <returns>
        /// Dictionary with objects for all the host controllers; the key is the controller ID
        /// </returns>
        public static Dictionary<string, SdkTbtController> GetControllersFromWmi()
        {
            // Assumes that the Thunderbolt service exists and can start running.
            Utilities.StartService();

            // ManagementObjectSearcher c-tor throws exception if SW isn't installed
            using (var mos = new ManagementObjectSearcher(@"root\Intel\Thunderbolt", "SELECT * FROM SdkTbtController"))
            {
                var controllers = new Dictionary<string, SdkTbtController>();
                foreach (var controller in mos.Get().Cast<ManagementObject>())
                {
                    var c = new SdkTbtController(controller);
                    controllers.Add(c.ControllerId, c);
                }
                return controllers;
            }

        }

        /// <summary>
        /// Validates FW image file compatibility with the current controller
        /// </summary>
        /// <param name="path">Path to FW image file to check</param>
        /// <remarks>
        /// When controller is in safe-mode, it's impossible to read its FW to gather needed info
        /// for comparison, so the function is no-op! User has to check if the controller is in
        /// safe-mode to know if validation had any effect.
        /// </remarks>
        /// <exception>Throws exception in case of incompatibility error or other errors</exception>
        public override void ValidateImage(string path)
        {
            if (!File.Exists(path))
            {
                throw new TbtException(TbtStatus.SDK_FILE_NOT_FOUND);
            }
            if (IsInSafeMode)
            {
                // No validation is possible; applications are advised to warn the user about it
                return;
            }
            using (var fs = new FileStream(path, FileMode.Open, FileAccess.Read))
            {
                var reader = new BinaryReader(fs);
                var imageBuffer = reader.ReadBytes((int) fs.Length);

                var controllerFwInfo = new HostFwInfo(new ControllerFwInfoSource(this));

                if (controllerFwInfo.Info.Generation == HwGeneration.DSL4510_4410)
                {
                    throw new TbtException(TbtStatus.SDK_REDWOOD_NOT_SUPPORTED);
                }

                var fileFwInfo = new HostFwInfo(new FileFwInfoSource(imageBuffer));
                if (controllerFwInfo.Info.Generation != fileFwInfo.Info.Generation)
                {
                    throw new TbtException(TbtStatus.SDK_HW_GENERATION_MISMATCH);
                }
                if (controllerFwInfo.Info.Type != fileFwInfo.Info.Type)
                {
                    throw new TbtException(TbtStatus.SDK_PORT_COUNT_MISMATCH);
                }

                var validator = new HostImageValidator(this,
                    imageBuffer,
                    controllerFwInfo.GetSectionInfo(),
                    fileFwInfo.GetSectionInfo(),
                    controllerFwInfo.Info);
                validator.Validate();
            }
        }

        /// <summary>
        /// See SdkTbtBase.GetCurrentNvmVersion().
        /// This function is here to throw in case of safe-mode
        /// </summary>
        public override UInt32 GetCurrentNvmVersion()
        {
            if (IsInSafeMode)
            {
                throw new TbtException(TbtStatus.SDK_INVALID_OPERATION_IN_SAFE_MODE);
            }
            return base.GetCurrentNvmVersion();
        }

        /// <summary>
        /// See SdkTbtBase.GetCurrentFullNvmVersion().
        /// This function is here to throw in case of safe-mode
        /// </summary>
        public override string GetCurrentFullNvmVersion()
        {
            if (IsInSafeMode)
            {
                throw new TbtException(TbtStatus.SDK_INVALID_OPERATION_IN_SAFE_MODE);
            }
            return base.GetCurrentFullNvmVersion();
        }

        /// <summary>
        /// See SdkTbtBase.GetCurrentPdVersion().
        /// This function is here to throw in case of safe-mode
        /// </summary>
        public override string GetCurrentPdVersion()
        {
            if (IsInSafeMode)
            {
                throw new TbtException(TbtStatus.SDK_INVALID_OPERATION_IN_SAFE_MODE);
            }
            return base.GetCurrentPdVersion();
        }

        /// <summary>
        /// See SdkTbtBase.ReadFirmware().
        /// This function is here to throw in case of safe-mode
        /// </summary>
        public override byte[] ReadFirmware(UInt32 offset, UInt32 length)
        {
            if (IsInSafeMode)
            {
                throw new TbtException(TbtStatus.SDK_INVALID_OPERATION_IN_SAFE_MODE);
            }
            return base.ReadFirmware(offset, length);
        }
    }
}
