﻿/*******************************************************************************
* Copyright (C) 2014 - 2016 Intel Corp. All rights reserved.
*
* Redistribution and use in source and binary forms, with or without
* modification, are permitted provided that the following conditions are met:
*
*  - Redistributions of source code must retain the above copyright notice,
*    this list of conditions and the following disclaimer.
*
*  - Redistributions in binary form must reproduce the above copyright notice,
*    this list of conditions and the following disclaimer in the documentation
*    and/or other materials provided with the distribution.
*
*  - Neither the name of Intel Corp. nor the names of its
*    contributors may be used to endorse or promote products derived from this
*    software without specific prior written permission.
*
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS ``AS IS''
* AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
* IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
* ARE DISCLAIMED. IN NO EVENT SHALL Intel Corp. OR THE CONTRIBUTORS
* BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
* CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
* SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
* CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
* ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
* POSSIBILITY OF SUCH DAMAGE.
*******************************************************************************/

using System;
using System.Collections.Generic;
using System.Linq;
using System.Management;
using System.IO;

namespace FwUpdateAPI
{
    /// <summary>
    /// This class is the interface to a device controller
    ///
    /// This, with its base class, is the main interface of this API module for applications
    /// to interface with device controllers.
    /// The main role of this class is to wrap the interface to WMI but it includes also the
    /// interface for validating image compatibility with the current controller.
    /// </summary>
    public class SdkTbtDevice : SdkTbtBase
    {
        /// <summary>
        /// C-tor
        /// </summary>
        /// <param name="mo">WMI object to use for interfacing with the controller</param>
        public SdkTbtDevice(ManagementObject mo)
            : base(mo)
        {
        }

        /// <summary>
        /// This property is used to differentiate between the power down messages
        /// given to user for host controller or device controller after update in
        /// cases where the PD version was updated. 
        /// </summary>
        /// <returns>Power down necessary for device controller message string</returns>
        public override string GetNeedPowerDownMessage
        {
            get { return Resources.DeviceNeedsPowerDownMessage; }
        }

        /// <summary>
        /// Device controller UUID property as returned by WMI
        /// </summary>
        public string UUID
        {
            get { return Mo.GetPropertyValue("UUID").ToString(); }
        }

        /// <summary>
        /// ID of the host controller the device is connected to as returned by WMI
        /// </summary>
        public string ControllerId
        {
            get { return Mo.GetPropertyValue("ControllerId").ToString(); }
        }

        /// <summary>
        /// 0-based index of the port in the host controller the device is connected to as
        /// returned by WMI
        /// </summary>
        public UInt32 PortNum
        {
            get { return (UInt32) Mo.GetPropertyValue("PortNum"); }
        }

        /// <summary>
        /// 1-based index of the device position in the port the device is connected to as
        /// returned by WMI
        /// </summary>
        public UInt32 PositionInChain
        {
            get { return (UInt32) Mo.GetPropertyValue("PositionInChain"); }
        }

        /// <summary>
        /// Device vendor name as returned by WMI
        /// </summary>
        public string VendorName
        {
            get { return Mo.GetPropertyValue("VendorName").ToString(); }
        }

        /// <summary>
        /// Device model name as returned by WMI
        /// </summary>
        public string ModelName
        {
            get { return Mo.GetPropertyValue("ModelName").ToString(); }
        }

        /// <summary>
        /// Device vendor ID as returned by WMI
        /// </summary>
        public UInt16 VendorId
        {
            // For some technical reasons the WMI is passing 32-bit number, but it's really 16-bit
            get { return (UInt16) (UInt32) Mo.GetPropertyValue("VendorId"); }
        }

        /// <summary>
        /// Device model ID as returned by WMI
        /// </summary>
        public UInt16 ModelId
        {
            // For some technical reasons the WMI is passing 32-bit number, but it's really 16-bit
            get { return (UInt16) (UInt32) Mo.GetPropertyValue("ModelId"); }
        }

        /// <summary>
        /// Controller enumeration out of n controllers in the device (as marked in the controller's NVM)
        /// </summary>
        public byte ControllerNum
        {
            get { return (byte)Mo.GetPropertyValue("ControllerNumber"); }
        }

        /// <summary>
        /// Total amount of controllers in the device
        /// </summary>
        public byte NumOfControllers
        {
            get { return (byte)Mo.GetPropertyValue("NumberOfControllers"); }
        }

        /// <summary>
        /// True if the device is updatable according to the information from WMI
        /// </summary>
        /// <remarks>
        /// If the controller doesn't support device FW update, the device appears like not
        /// updatable even if it supports FW update.
        /// </remarks>
        public bool Updatable
        {
            get { return (bool) Mo.GetPropertyValue("Updatable"); }
        }

        /// <summary>
        /// Gets list of SdkTbtController objects for each of the device controllers exposed by WMI
        /// </summary>
        /// <returns>
        /// Dictionary with objects for all the device controllers; the key is the device UUID
        /// </returns>
        public static Dictionary<string, SdkTbtDevice> GetDevicesFromWmi()
        {
            // Assumes that the Thunderbolt service exists and can start running.
            Utilities.StartService();

            // ManagementObjectSearcher c-tor throws exception if SW isn't installed
            using (var mos = new ManagementObjectSearcher(@"root\Intel\Thunderbolt", "SELECT * FROM SdkTbtDevice"))
            {
                var devices = new Dictionary<string, SdkTbtDevice>();
                foreach (var controller in mos.Get().Cast<ManagementObject>())
                {
                    var c = new SdkTbtDevice(controller);
                    devices.Add(c.UUID, c);
                }
                return devices;
            }
        }

        /// <summary>
        /// Validates FW image file compatibility with the current controller
        /// </summary>
        /// <param name="path">Path to FW image file to check</param>
        /// <exception>Throws exception in case of incompatibility error or other errors</exception>
        public override void ValidateImage(string path)
        {
            if (!File.Exists(path))
            {
                throw new TbtException(TbtStatus.SDK_FILE_NOT_FOUND);
            }
            using (var fs = new FileStream(path, FileMode.Open, FileAccess.Read))
            {
                var reader = new BinaryReader(fs);
                var imageBuffer = reader.ReadBytes((int) fs.Length);

                var controller = SdkTbtController.GetControllersFromWmi()[ControllerId];

                if (new ControllerFwInfoSource(controller).Info.Generation < HwGeneration.DSL6540_6340)
                {
                    throw new TbtException(TbtStatus.SDK_NO_EP_UPDATE_SUPPORT);
                }

                var deviceFwInfo = new DeviceFwInfo(new ControllerFwInfoSource(this));
                if (deviceFwInfo.Info.Generation < HwGeneration.DSL6540_6340)
                {
                    throw new TbtException(TbtStatus.SDK_DEVICE_NOT_SUPPORTED);
                }

                var fileFwInfo = new DeviceFwInfo(new FileFwInfoSource(imageBuffer));
                if (deviceFwInfo.Info.Generation != fileFwInfo.Info.Generation)
                {
                    throw new TbtException(TbtStatus.SDK_HW_GENERATION_MISMATCH);
                }
                if (deviceFwInfo.Info.Type != fileFwInfo.Info.Type)
                {
                    throw new TbtException(TbtStatus.SDK_PORT_COUNT_MISMATCH);
                }

                var validator = new DeviceImageValidator(this,
                    imageBuffer,
                    deviceFwInfo.GetSectionInfo(),
                    fileFwInfo.GetSectionInfo(),
                    deviceFwInfo.Info);
                validator.Validate();
            }
        }
    }
}
